from pulpo.utils import optimizer, bw_parser, converter, saver
from typing import List, Union
import webbrowser
from tests.rice_database import setup_rice_husk_db
from tests.generic_database import setup_generic_db

class PulpoOptimizer:
    def __init__(self, project: str, database: str, method: Union[str, List[str], dict], directory: str):
        """
        Initializes the PulpoOptimizer with project, database, method, and directory.

        Args:
            project (str): Name of the project.
            database (str): Name of the database.
            method (Union[str, List[str], dict]): Method(s) for optimization.
            directory (str): Directory for saving results.
        """
        self.project = project
        self.database = database
        self.intervention_matrix = 'biosphere3'
        self.method = converter.convert_to_dict(method)
        self.directory = directory
        self.lci_data = None
        self.instance = None

    def get_lci_data(self):
        """
        Imports LCI data for the project using the specified database and method.
        """
        self.lci_data = bw_parser.import_data(self.project, self.database, self.method, self.intervention_matrix)

    def instantiate(self, choices={}, demand={}, upper_limit={}, lower_limit={}, upper_elem_limit={},
                    upper_imp_limit={}):
        """
        Combines inputs and instantiates the optimization model.

        Args:
            choices (dict): Choices for the model.
            demand (dict): Demand data.
            upper_limit (dict): Upper limit constraints.
            lower_limit (dict): Lower limit constraints.
            upper_elem_limit (dict): Upper elemental limit constraints.
            upper_imp_limit (dict): Upper impact limit constraints.
        """
        # Instantiate only for those methods that are part of the objective or the limits
        methods = {h: self.method[h] for h in self.method if self.method[h] != 0 or h in upper_imp_limit}
        data = converter.combine_inputs(self.lci_data, demand, choices, upper_limit, lower_limit, upper_elem_limit,
                                        upper_imp_limit, methods)
        self.instance = optimizer.instantiate(data)

    def solve(self, GAMS_PATH=False, solver_name=None, options=None):
        """
        Solves the optimization model and calculates additional methods and inventory flows if needed.

        Args:
            GAMS_PATH (bool): Path to GAMS if needed.
            options (dict): Additional options for the solver.

        Returns:
            results: Results of the optimization.
        """
        results, self.instance = optimizer.solve_model(self.instance, GAMS_PATH, solver_name=solver_name, options=options)

        # Post calculate additional methods, in case several methods have been specified and one of them is 0
        if not isinstance(self.method, str):
            if len(self.method) > 1 and 0 in [self.method[x] for x in self.method]:
                self.instance = optimizer.calculate_methods(self.instance, self.lci_data, self.method)

        self.instance = optimizer.calculate_inv_flows(self.instance, self.lci_data)
        return results

    def retrieve_processes(self, keys=None, processes=None, reference_products=None, locations=None):
        """
        Retrieves processes from the database based on given filters.

        Args:
            keys (list): List of keys to filter activities.
            processes (list): List of processes to filter.
            reference_products (list): List of reference products to filter.
            locations (list): List of locations to filter.

        Returns:
            activities: Filtered activities from the database.
        """
        processes = bw_parser.retrieve_processes(self.project, self.database, keys, processes, reference_products,
                                                  locations)
        return processes

    def retrieve_activities(self, keys=None, activities=None, reference_products=None, locations=None):
        """
        Works the same as "retrieve_processes" but with a different name. Will be obsolete in future versions.
        """
        activities = bw_parser.retrieve_processes(self.project, self.database, keys, activities, reference_products,
                                                  locations)
        return activities

    def retrieve_envflows(self, keys=None, activities=None, categories=None):
        """
        Retrieves environmental flows from the database based on given filters.

        Args:
            keys (list): List of keys to filter environmental flows.
            activities (list): List of activities to filter.
            categories (list): List of categories to filter.

        Returns:
            activities: Filtered environmental flows from the database.
        """
        activities = bw_parser.retrieve_env_interventions(project=self.project,
                                                          intervention_matrix=self.intervention_matrix, keys=keys,
                                                          activities=activities, categories=categories)
        return activities

    def retrieve_methods(self, string=""):
        """
        Retrieves methods from the database based on a search string.

        Args:
            string (str): Search string for methods.

        Returns:
            methods: List of methods that match the search string.
        """
        methods = bw_parser.retrieve_methods(self.project, string)
        return methods

    def save_results(self, choices={}, constraints={}, demand={}, name='results.xlxs'):
        """
        Saves the results of the optimization to a file.

        Args:
            choices (dict): Choices for the model.
            constraints (dict): Constraints applied during optimization.
            demand (dict): Demand data used in optimization.
            name (str): Name of the file to save results.
        """
        saver.save_results(self.instance, self.project, self.database, choices, constraints, demand,
                           self.lci_data['process_map'], self.lci_data['intervention_map'], self.directory, name)

    def summarize_results(self, choices={}, constraints={}, demand={}, zeroes=False):
        """
        Summarizes the results of the optimization.

        Args:
            choices (dict): Choices for the model.
            constraints (dict): Constraints applied during optimization.
            demand (dict): Demand data used in optimization.
            zeroes (bool): Whether to include zero values in the summary.
        """
        saver.summarize_results(self.instance, choices, constraints, demand,
                                self.lci_data['process_map'], zeroes)


def electricity_showcase():
    """
    Opens the electricity showcase notebook in the web browser.
    """
    github_url = 'https://github.com/flechtenberg/pulpo/blob/master/notebooks/electricity_showcase.ipynb'
    nbviewer_url = 'https://nbviewer.jupyter.org/github/' + github_url.split('github.com/')[1]
    webbrowser.open(nbviewer_url)


def hydrogen_showcase():
    """
    Opens the hydrogen showcase notebook in the web browser.
    """
    github_url = 'https://github.com/flechtenberg/pulpo/blob/master/notebooks/hydrogen_showcase.ipynb'
    nbviewer_url = 'https://nbviewer.jupyter.org/github/' + github_url.split('github.com/')[1]
    webbrowser.open(nbviewer_url)


def plastic_showcase():
    """
    Opens the plastic showcase notebook in the web browser.
    """
    github_url = 'https://github.com/flechtenberg/pulpo/blob/master/notebooks/plastic_showcase.ipynb'
    nbviewer_url = 'https://nbviewer.jupyter.org/github/' + github_url.split('github.com/')[1]
    webbrowser.open(nbviewer_url)


def install_rice_husk_db():
    """
    Sets up the rice husk example database.
    """
    setup_rice_husk_db()

def install_generic_db(project="generic_db_project", database="generic_db", n_prod=5, n_proc=3, n_reg=3, n_inputs=4, n_flows=4, n_methods=2, seed=None, return_data=False):
    """
    Sets up the generic LCI database in Brightway2 with specified parameters.

    Args:
        project (str): Name of the Brightway2 project to create or use. Defaults to "generic_db_project".
        database (str): Name of the database to create or use. Defaults to "generic_db".
        n_prod (int): Number of products to generate. Defaults to 5.
        n_proc (int): Maximum number of processes per product. Defaults to 3.
        n_reg (int): Number of regions where processes can be active. Defaults to 3.
        n_inputs (int): Maximum number of inputs per process. Defaults to 4.
        n_flows (int): Number of environmental flows to generate. Defaults to 4.
        n_methods (int): Number of impact assessment methods to create. Defaults to 2.
        seed (int, optional): Seed for reproducibility of random data generation. Defaults to None.
        return_data (bool): If True, returns the generated matrices (technosphere, biosphere, and
            characterization). Defaults to False.

    Returns:
        tuple: If `return_data` is True, returns a tuple containing:
            - technosphere_matrix (np.ndarray): The technosphere matrix.
            - biosphere_matrix (np.ndarray): The biosphere matrix.
            - characterization_matrices (dict): A dictionary of characterization factor matrices.
    """
    return setup_generic_db(project=project, database=database, n_prod=n_prod, n_proc=n_proc, n_reg=n_reg,
                            n_inputs=n_inputs, n_flows=n_flows, n_methods=n_methods,
                            seed=seed, return_data=return_data)




from typing import List, Union, Dict, Any
import brightway2 as bw

def import_data(project: str, database: str, method: Union[str, List[str], dict[str, int]],
                intervention_matrix_name: str) -> Dict[str, Union[dict, Any]]:
    """
    Main function to import LCI data for a project from a database.

    Args:
        project (str): Name of the project.
        database (str): Name of the database.
        method (Union[str, List[str], dict[str, int]]): Method(s) for data retrieval.
        intervention_matrix_name (str): Name of the intervention matrix.

    Returns:
        Dict[str, Union[dict, Any]]: Dictionary containing imported LCI data.
    """

    # Set project and get database
    bw.projects.set_current(project)
    eidb = bw.Database(database)

    # Prepare methods
    if isinstance(method, str):
        method = [method]  # Convert single string to list of strings
    method = sorted(method)
    methods = retrieve_methods(project, method)

    # Get data
    characterization_matrices = {}
    rand_act = eidb.random()
    for method in methods:
        lca = bw.LCA({rand_act: 1}, method)
        lca.load_lci_data()
        lca.load_lcia_data()
        characterization_matrices[str(method)] = lca.characterization_matrix

    technology_matrix = lca.technosphere_matrix
    intervention_matrix = lca.biosphere_matrix

    # Create activity map key --> ID | ID --> description
    process_map = lca.product_dict
    for act in eidb:
        process_map[process_map[act.key]] = str(act['name']) + ' | ' + str(act['reference product']) + ' | ' + str(
            act['location'])

    if intervention_matrix_name in bw.databases:
        eidb_bio = bw.Database(intervention_matrix_name)
        intervention_map = lca.biosphere_dict
        for act in eidb_bio:
            if act.key in intervention_map:
                intervention_map[intervention_map[act.key]] = act['name'] + ' | ' + str(act['categories'])
    else:
        print(
            "The name of the biosphere is not '" + intervention_matrix_name + "'. Please specify the correct biosphere.")
        return {}

    lci_data = {
        'matrices': characterization_matrices,
        'intervention_matrix': intervention_matrix,
        'technology_matrix': technology_matrix,
        'process_map': process_map,
        'intervention_map': intervention_map,
    }

    return lci_data


def retrieve_processes(project: str, database: str, keys=None, activities=None, reference_products=None,
                       locations=None):
    """
    Retrieve activities from a database based on specified keys, activities, reference products, and locations.

    Args:
        project (str): Name of the project.
        database (str): Name of the database.
        keys (list, optional): List of keys to filter activities.
        activities (list, optional): List of activity names to filter.
        reference_products (list, optional): List of reference products to filter.
        locations (list, optional): List of locations to filter.

    Returns:
        list: List of matching activities from the database.
    """

    # Set project and get database
    bw.projects.set_current(project)
    eidb = bw.Database(database)

    # Filter by keys if provided
    if keys is not None:
        if isinstance(keys, str):
            keys = [keys]
        keys = [eval(key) for key in keys]
        return [proc for proc in eidb if proc.key in keys]

    matching_processes = []

    # Filter by activities, reference products, and locations
    for proc in eidb:
        if (activities is None or proc['name'] in activities) and \
                (reference_products is None or proc['reference product'] in reference_products) and \
                (locations is None or proc['location'] in locations):
            matching_processes.append(proc)

    if not matching_processes:
        print("No activities match the given specifications or the input format is incorrect.")
    else:
        return matching_processes


def retrieve_env_interventions(project: str = '', intervention_matrix: str = 'biosphere3', keys=None, activities=None,
                               categories=None):
    """
    Retrieve environmental flows from the biosphere database based on specified keys, activities, and categories.

    Args:
        project (str, optional): Name of the project.
        intervention_matrix (str): Name of the intervention matrix.
        keys (list, optional): List of keys to filter environmental flows.
        activities (list, optional): List of activity names to filter.
        categories (list, optional): List of categories to filter.

    Returns:
        list: List of matching environmental flows from the database.
    """

    # Set project and get database
    bw.projects.set_current(project)
    eidb = bw.Database(intervention_matrix)

    # Filter by keys if provided
    if keys is not None:
        if isinstance(keys, str):
            keys = [keys]
        keys = [eval(key) for key in keys]
        return [flow for flow in eidb if flow.key in keys]

    matching_flows = []

    # Filter by activities and categories
    for flow in eidb:
        if (activities is None or flow['name'] in activities) and \
                (categories is None or str(flow['categories']) in categories):
            matching_flows.append(flow)

    if not matching_flows:
        print("No flows match the given specifications or the input format is incorrect.")
    else:
        return matching_flows


def retrieve_methods(project: str, sub_string: List[str]) -> List[str]:
    """
    Retrieve all methods that contain the specified list of substrings.

    Args:
        project (str): Name of the project.
        sub_string (List[str]): List of substrings to search for in method names.

    Returns:
        List[str]: List of methods that match the substrings.
    """
    bw.projects.set_current(project)
    return [method for method in bw.methods if any([x.lower() in str(method).lower() for x in sub_string])]



from brightway2 import Database, projects, Method, databases, methods
import brightway2 as bw
import copy
import pulpo as pulpo
from pulpo.utils.bw_parser import import_data, retrieve_methods, retrieve_env_interventions, retrieve_processes
from pulpo import pulpo

def setup_test_db():
    # Set the current project to "sample_project"
    projects.set_current("sample_project")

    # Keys
    # Biosphere
    co2_key = ('biosphere', 'CO2')
    ch4_key = ('biosphere', 'CH4')
    pm_key = ('biosphere', 'PM')
    h2o_irrigation_key = ('biosphere', 'H2O_irrigation')
    # Technosphere
    wind_turbine_key = ('technosphere', 'wind turbine')
    steam_cycle_key = ('technosphere', 'steam cycle')
    lignite_extraction_key = ('technosphere', 'lignite extraction')
    oil_extraction_key = ('technosphere', 'oil extraction')
    e_car_key = ('technosphere', 'e-Car')

    # Define biosphere flows and create the "biosphere" database if it doesn't exist
    if "biosphere" not in databases:
        biosphere_db = Database("biosphere")
        biosphere_data = {
            ("biosphere", "CO2"): {
                "name": "Carbon dioxide, fossil",
                "categories": ("climate change", "GWP 100a"),
                "type": "emission",
                "unit": "kg",
            },
            ("biosphere", "CH4"): {
                "name": "Methane, agricultural",
                "categories": ("climate change", "GWP 100a"),
                "type": "emission",
                "unit": "kg",
            },
            ("biosphere", "PM"): {
                "name": "Particulate matter, industrial",
                "categories": ("air quality", "particulate matter"),
                "type": "emission",
                "unit": "g",
            },
            ("biosphere", "H2O_irrigation"): {
                "name": "Water, irrigation",
                "categories": ("water use", "irrigation"),
                "type": "resource",
                "unit": "m3",
            },
        }
        biosphere_db.write(biosphere_data)

    # Create the "technosphere" database if it doesn't exist
    if "technosphere" not in databases:
        technosphere_db = Database("technosphere")
        technosphere_db.write({})
        process_data = [
            ("oil extraction", "kg", "GLO", "oil"),
            ("lignite extraction", "kg", "GLO", "lignite"),
            ("steam cycle", "kWh", "GLO", "electricity"),
            ("wind turbine", "kWh", "GLO", "electricity"),
            ("e-Car", "tkm", "GLO", "transport"),
        ]

        for name, unit, location, ref_product in process_data:
            act = technosphere_db.new_activity(name)
            act["unit"] = unit
            act["location"] = location
            act["name"] = name
            act["reference product"] = ref_product
            act.new_exchange(amount=1.0, input=act.key, type="production").save()
            act.save()

        exchange_data = [
            [e_car_key, oil_extraction_key, 0.03, 'technosphere'],
            [e_car_key, lignite_extraction_key, 0.03, 'technosphere'],
            [oil_extraction_key, steam_cycle_key, 0.5, 'technosphere'],
            [lignite_extraction_key, steam_cycle_key, 0.5, 'technosphere'],
            [steam_cycle_key, e_car_key, 0.5, 'technosphere'],
            [wind_turbine_key, e_car_key, 0.5, 'technosphere'],
            [e_car_key, wind_turbine_key, 0.03, 'technosphere'],
            [co2_key, oil_extraction_key, 0.2, 'biosphere'],
            [co2_key, lignite_extraction_key, 0.3, 'biosphere'],
            [co2_key, steam_cycle_key, 1, 'biosphere'],
            [co2_key, wind_turbine_key, 0.1, 'biosphere'],
            [ch4_key, oil_extraction_key, 0.01, 'biosphere'],
            [ch4_key, lignite_extraction_key, 0.02, 'biosphere'],
            [pm_key, steam_cycle_key, 2.0, 'biosphere'],
            [pm_key, wind_turbine_key, 1.5, 'biosphere'],
            [h2o_irrigation_key, steam_cycle_key, 2.0, 'biosphere'],
            [h2o_irrigation_key, wind_turbine_key, 5.0, 'biosphere'],
            [pm_key, e_car_key, 0.1, 'biosphere'],
            [h2o_irrigation_key, e_car_key, 0.1, 'biosphere'],
        ]

        for input, target, amount, type in exchange_data:
            act = [act for act in technosphere_db if act.key==target][0]
            act.new_exchange(amount=amount, input=input, type=type).save()
            act.save()

    # Loop through the list of methods and deregister each one
    # Create a copy of the methods list
    methods_copy = copy.deepcopy(methods)
    # Loop through the copy of methods and deregister each one
    for method in methods_copy:
        Method(method).deregister()

    # Define LCIA methods and CFs
    methods_data = [
        ("climate change", "kg CO2eq", 2, "cc", "climate change CFs", "climate_change", "CO2", [(co2_key, 1), (ch4_key, 29.7)]),
        ("air quality", "ppm", 1, "aq", "air quality CFs", "air_quality", "PM", [(ch4_key, 29.7)]),
        ("resources", "m3", 1, "rc", "resource CFs", "resources", "H2O_irrigation", [(h2o_irrigation_key,1)]),
    ]

    for method_name, unit, num_cfs, abbreviation, description, filename, flow_code, flow_list in methods_data:
        method = Method(("my project", method_name))
        method.register(**{
            "unit": unit,
            "num_cfs": num_cfs,
            "abbreviation": abbreviation,
            "description": description,
            "filename": filename,
        })
        method.write(flow_list)

def sample_lcia():
    # Load the technosphere database
    bw.projects.set_current('sample_project')
    technosphere_db = bw.Database("technosphere")
    # Perform LCA with FU 1 for all activities
    act = {act.key: 1 for act in technosphere_db}
    # Perform Multi-LCA
    bw.calculation_setups['multiLCA'] = {'inv': [act], 'ia': list(methods)}
    myMultiLCA = bw.MultiLCA('multiLCA')
    results = [round(x, 5) for x in myMultiLCA.results[0]]
    return results

def test_database_import():
    result = sample_lcia()
    assert result == [4.22308, 1.5937, 11.1567]  # Example assertion

def test_import_data():
    # Test if the import works and has the expected structure
    methods = {
        "('my project', 'climate change')": 1,
        "('my project', 'air quality')": 1,
        "('my project', 'resources')": 1,
    }

    result = import_data('sample_project', 'technosphere', methods, 'biosphere')

    assert [idx for idx in result] == ['matrices', 'intervention_matrix', 'technology_matrix', 'process_map', 'intervention_map']
    assert result['technology_matrix'].shape == (5, 5)
    assert result['intervention_matrix'].shape == (4, 5)
    assert result['process_map'] == {('technosphere', 'oil extraction'): 0, ('technosphere', 'lignite extraction'): 1, ('technosphere', 'steam cycle'): 2, ('technosphere', 'wind turbine'): 3, ('technosphere', 'e-Car'): 4, 2: 'steam cycle | electricity | GLO', 1: 'lignite extraction | lignite | GLO', 0: 'oil extraction | oil | GLO', 3: 'wind turbine | electricity | GLO', 4: 'e-Car | transport | GLO'}

def test_retrieve_activities():
    key = retrieve_processes('sample_project', 'technosphere', keys=["('technosphere', 'wind turbine')"])
    name = retrieve_processes('sample_project', 'technosphere', activities=['e-Car'])
    location = retrieve_processes('sample_project', 'technosphere', locations=['GLO'])
    assert key[0]['name'] == "wind turbine"
    assert name[0]['name'] == "e-Car"
    assert len(location) == 5

def test_retrieve_methods():
    single_result = retrieve_methods('sample_project', ['climate'])
    multi_result = retrieve_methods('sample_project', ['project'])
    assert single_result == [('my project', 'climate change')]
    assert multi_result == [('my project', 'climate change'), ('my project', 'air quality'), ('my project', 'resources')]

def test_retrieve_envflows():
    result = retrieve_env_interventions('sample_project', intervention_matrix='biosphere', keys="('biosphere', 'PM')")
    assert result[0]['name'] == 'Particulate matter, industrial'

def test_pulpo():
    project = 'sample_project'
    database = 'technosphere'
    methods = {"('my project', 'climate change')": 1,
               "('my project', 'air quality')": 1,
               "('my project', 'resources')": 0}

    worker = pulpo.PulpoOptimizer(project, database, methods, '')
    worker.intervention_matrix = 'biosphere'
    worker.get_lci_data()
    eCar = worker.retrieve_activities(reference_products='transport')
    demand = {eCar[0]: 1}
    elec = worker.retrieve_activities(reference_products='electricity')
    choices = {'electricity': {elec[0]: 100, elec[1]: 100}}
    worker.instantiate(choices=choices, demand=demand)
    worker.solve()
    result_obj = round(worker.instance.OBJ(), 6)
    result_aux = round(worker.instance.impacts_calculated["('my project', 'resources')"].value, 5)
    assert result_obj == 0.103093
    assert result_aux == 5.25773

    upper_limit = {eCar[0]: 1}
    lower_limit = {eCar[0]: 1}
    worker.instantiate(choices=choices, upper_limit=upper_limit, lower_limit=lower_limit)
    worker.solve()
    result_obj = round(worker.instance.OBJ(), 6)
    result_aux = round(worker.instance.impacts_calculated["('my project', 'resources')"].value, 5)
    assert result_obj == 0.1
    assert result_aux == 5.1

def setup():
    test_import_data()
    test_database_import()
    test_retrieve_activities()
    test_retrieve_envflows()
    test_retrieve_methods()
    test_pulpo()
    print('\nAll tests passed successfully. You are good to go!')


import brightway2 as bw
import copy
import numpy as np

np.NaN = np.nan  # Ensures compatibility with the latest NumPy versions

def setup_rice_husk_db():
    # Set the current project to "rice_husk_example"
    bw.projects.set_current("rice_husk_example")

    # Biosphere keys
    co2_key = ('biosphere3', 'CO2')
    ch4_key = ('biosphere3', 'CH4')
    ef_key = ('biosphere3', 'EF')

    # Define biosphere flows and create the 'biosphere3' database if it doesn't exist
    if 'biosphere3' not in bw.databases:
        biosphere_db = bw.Database('biosphere3')
        biosphere_data = {
            ('biosphere3', 'CO2'): {
                'name': 'Carbon dioxide, fossil',
                'categories': ('climate change', 'GWP 100a'),
                'type': 'emission',
                'unit': 'Mt',
            },
            ('biosphere3', 'CH4'): {
                'name': 'Methane, agricultural',
                'categories': ('climate change', 'GWP 100a'),
                'type': 'emission',
                'unit': 'Mt',
            },
            ('biosphere3', 'EF'): {
                'name': 'Economic Flow',
                'categories': ('economic flow', 'dollar'),
                'type': 'emission',
                'unit': 'million dollar',
            },
        }
        biosphere_db.write(biosphere_data)

        print('Elementary flow database created')

    # Technosphere keys
    rice_factory_key = ('rice_husk_example_db', 'Rice factory')
    rice_farming_key = ('rice_husk_example_db', 'Rice farming')
    rice_husk_boiler_key = ('rice_husk_example_db', 'Rice husk boiler')
    natural_gas_boiler_key = ('rice_husk_example_db', 'Natural gas boiler')
    wood_pellet_boiler_key = ('rice_husk_example_db', 'Wood pellet boiler')
    rice_husk_market_key = ('rice_husk_example_db', 'Rice husk market')
    rice_husk_collection1_key = ('rice_husk_example_db', 'Rice husk collection 1')
    rice_husk_collection2_key = ('rice_husk_example_db', 'Rice husk collection 2')
    rice_husk_collection3_key = ('rice_husk_example_db', 'Rice husk collection 3')
    rice_husk_collection4_key = ('rice_husk_example_db', 'Rice husk collection 4')
    rice_husk_collection5_key = ('rice_husk_example_db', 'Rice husk collection 5')
    natural_gas_supply_key = ('rice_husk_example_db', 'Natural gas supply')
    wood_pellet_supply_key = ('rice_husk_example_db', 'Wood pellet supply')
    burning_rice_husk_key = ('rice_husk_example_db', 'Burning of rice husk')
    power_plant_key = ('rice_husk_example_db', 'Power plant')
    transportation_key = ('rice_husk_example_db', 'Transportation by truck')

    # Create the "rice_husk_example_db" database if it doesn't exist
    if 'rice_husk_example_db' not in bw.databases:
        rice_husk_db = bw.Database('rice_husk_example_db')
        rice_husk_db.write({})

        process_data = [
            ('Rice factory', 'Mt', 'GLO', 'Processed rice (in Mt)'),
            ('Rice farming', 'Mt', 'GLO', 'Unprocessed rice (in Mt)'),
            ('Rice husk boiler', 'TWh', 'GLO', 'Thermal energy from husk boiler (in TWh)'),
            ('Natural gas boiler', 'TWh', 'GLO', 'Thermal energy from natural gas boiler (in TWh)'),
            ('Wood pellet boiler', 'TWh', 'GLO', 'Thermal energy from wood pellet boiler (in TWh)'),
            ('Rice husk market', 'Mt', 'GLO', 'Rice husk at farm (Mt)'),
            ('Rice husk collection 1', 'Mt', 'GLO', 'Rice husk from region 1 (in Mt)'),
            ('Rice husk collection 2', 'Mt', 'GLO', 'Rice husk from region 2 (in Mt)'),
            ('Rice husk collection 3', 'Mt', 'GLO', 'Rice husk from region 3 (in Mt)'),
            ('Rice husk collection 4', 'Mt', 'GLO', 'Rice husk from region 4 (in Mt)'),
            ('Rice husk collection 5', 'Mt', 'GLO', 'Rice husk from region 5 (in Mt)'),
            ('Natural gas supply', 'TWh', 'GLO', 'Natural gas (in TWh)'),
            ('Wood pellet supply', 'Mt', 'GLO', 'Wood pellets (in Mt)'),
            ('Burning of rice husk', 'Mt', 'GLO', 'Burned rice husk (in Mt)'),
            ('Power plant', 'TWh', 'GLO', 'Electricity (in TWh)'),
            ('Transportation by truck', 'Gt*km', 'GLO', 'Transportation (in Gt*km)'),
        ]

        # Register each process
        for name, unit, location, ref_product in process_data:
            act = rice_husk_db.new_activity(name)
            act['unit'] = unit
            act['location'] = location
            act['name'] = name
            act['reference product'] = ref_product
            act.new_exchange(amount=1.0, input=act.key, type='production').save()
            act.save()

        # Define the technosphere exchanges
        exchange_data = [
            # Format: [input_key, target_key, amount, exchange_type]
            [rice_farming_key, rice_factory_key, 1.15, 'technosphere'],
            [natural_gas_boiler_key, rice_factory_key, 2.2, 'technosphere'],
            [power_plant_key, rice_factory_key, 0.08, 'technosphere'],
            [transportation_key, rice_factory_key, 0.35, 'technosphere'],
            [rice_husk_market_key, rice_farming_key, -0.6, 'technosphere'],
            [rice_husk_collection1_key, rice_husk_boiler_key, 0.23, 'technosphere'],
            [natural_gas_supply_key, natural_gas_boiler_key, 1.1, 'technosphere'],
            [wood_pellet_supply_key, wood_pellet_boiler_key, 0.25, 'technosphere'],
            [rice_husk_market_key, rice_husk_collection1_key, 1, 'technosphere'],
            [rice_husk_market_key, rice_husk_collection2_key, 1, 'technosphere'],
            [rice_husk_market_key, rice_husk_collection3_key, 1, 'technosphere'],
            [rice_husk_market_key, rice_husk_collection4_key, 1, 'technosphere'],
            [rice_husk_market_key, rice_husk_collection5_key, 1, 'technosphere'],
            [transportation_key, rice_husk_collection1_key, 0.12, 'technosphere'],
            [transportation_key, rice_husk_collection2_key, 0.24, 'technosphere'],
            [transportation_key, rice_husk_collection3_key, 0.36, 'technosphere'],
            [transportation_key, rice_husk_collection4_key, 0.48, 'technosphere'],
            [transportation_key, rice_husk_collection5_key, 0.60, 'technosphere'],
            [rice_husk_market_key, burning_rice_husk_key, 2, 'technosphere'],
            [burning_rice_husk_key, rice_husk_market_key, 1, 'technosphere'],
            [co2_key, rice_farming_key, 6.14e-1, 'biosphere'],
            [ch4_key, rice_farming_key, 1.33e-3, 'biosphere'],
            [co2_key, natural_gas_boiler_key, 2.27e-1, 'biosphere'],
            [ch4_key, natural_gas_boiler_key, 1.47e-3, 'biosphere'],
            [co2_key, natural_gas_supply_key, 3.21e-2, 'biosphere'],
            [ch4_key, natural_gas_supply_key, 1.50e-3, 'biosphere'],
            [co2_key, wood_pellet_supply_key, 1.50e-1, 'biosphere'],
            [ch4_key, wood_pellet_supply_key, 2.56e-4, 'biosphere'],
            [co2_key, power_plant_key, 1.10e-0, 'biosphere'],
            [ch4_key, power_plant_key, 9.15e-4, 'biosphere'],
            [co2_key, transportation_key, 5.76e-2, 'biosphere'],
            [ch4_key, transportation_key, 6.97e-5, 'biosphere'],
            [ef_key, rice_factory_key, 5.00e1, 'biosphere'],
            [ef_key, rice_farming_key, 3.60e2, 'biosphere'],
            [ef_key, natural_gas_supply_key, 1.3e1, 'biosphere'],
            [ef_key, wood_pellet_supply_key, 7.20e1, 'biosphere'],
            [ef_key, rice_husk_collection1_key, 4.50e1, 'biosphere'],
            [ef_key, rice_husk_collection2_key, 3.60e1, 'biosphere'],
            [ef_key, rice_husk_collection3_key, 2.90e1, 'biosphere'],
            [ef_key, rice_husk_collection4_key, 2.30e1, 'biosphere'],
            [ef_key, rice_husk_collection5_key, 1.80e1, 'biosphere'],
            [ef_key, power_plant_key, 6.50e1, 'biosphere'],
            [ef_key, transportation_key, 1.70e2, 'biosphere'],
        ]

        # Add the exchanges to the activities
        for input_key, target_key, amount, exchange_type in exchange_data:
            act = [act for act in rice_husk_db if act.key == target_key][0]
            act.new_exchange(amount=amount, input=input_key, type=exchange_type).save()
            act.save()

        print('Process database created')

        # Loop through the list of methods and deregister each one
        # Create a copy of the methods list
        methods_copy = copy.deepcopy(bw.methods)
        # Loop through the copy of methods and deregister each one
        for method in methods_copy:
            bw.Method(method).deregister()

        # Define LCIA methods and CFs
        methods_data = [
            ('climate change', 'Mt CO2eq', 2, 'cc', 'climate change CFs', 'climate_change', 'CO2',
             [(co2_key, 1), (ch4_key, 25)]),
            ('air quality', 'ppm', 1, 'aq', 'air quality CFs', 'air_quality', 'PM', [(ch4_key, 25)]),
            ('economic flow', 'million dollar', 1, 'ef', 'economic flow CFs', 'economic_flow', 'million dollar', [(ef_key, 1)]),
        ]

        for method_name, unit, num_cfs, abbreviation, description, filename, flow_code, flow_list in methods_data:
            method = bw.Method(('my project', method_name))
            method.register(**{
                'unit': unit,
                'num_cfs': num_cfs,
                'abbreviation': abbreviation,
                'description': description,
                'filename': filename,
            })
            method.write(flow_list)

        print('LCIA methods and CFs defined')

def main():
    setup_rice_husk_db()

if __name__ == '__main__':
    main()



import pulpo as pulpo
import unittest
from tests.sample_database import sample_lcia, setup_test_db
from pulpo.utils.bw_parser import import_data, retrieve_methods, retrieve_env_interventions, retrieve_processes
from pulpo import pulpo

setup_test_db()

class TestParser(unittest.TestCase):
    def test_database_import(self):
        result = sample_lcia()
        self.assertEqual(result, [4.22308, 1.5937, 11.1567])  # Example assertion

    def test_import_data(self):
        # Test if the import works and has the expected structure
        methods = {
            "('my project', 'climate change')": 1,
            "('my project', 'air quality')": 1,
            "('my project', 'resources')": 1,
        }

        result = import_data('sample_project', 'technosphere', methods, 'biosphere')

        self.assertEqual([idx for idx in result], ['matrices', 'intervention_matrix', 'technology_matrix', 'process_map', 'intervention_map'])
        self.assertEqual(result['technology_matrix'].shape, (5, 5))
        self.assertEqual(result['intervention_matrix'].shape, (4, 5))
        self.assertEqual(result['process_map'], {('technosphere', 'oil extraction'): 0, ('technosphere', 'lignite extraction'): 1, ('technosphere', 'steam cycle'): 2, ('technosphere', 'wind turbine'): 3, ('technosphere', 'e-Car'): 4, 2: 'steam cycle | electricity | GLO', 1: 'lignite extraction | lignite | GLO', 0: 'oil extraction | oil | GLO', 3: 'wind turbine | electricity | GLO', 4: 'e-Car | transport | GLO'})

    def test_retrieve_activities(self):
        key = retrieve_processes('sample_project', 'technosphere', keys=["('technosphere', 'wind turbine')"])
        name = retrieve_processes('sample_project', 'technosphere', activities=['e-Car'])
        location = retrieve_processes('sample_project', 'technosphere', locations=['GLO'])
        self.assertEqual(key[0]['name'], "wind turbine")
        self.assertEqual(name[0]['name'], "e-Car")
        self.assertEqual(len(location), 5)

    def test_retrieve_methods(self):
        single_result = retrieve_methods('sample_project', ['climate'])
        multi_result = retrieve_methods('sample_project', ['project'])
        self.assertEqual(single_result, [('my project', 'climate change')])
        self.assertEqual(multi_result, [('my project', 'climate change'), ('my project', 'air quality'), ('my project', 'resources')])

    def test_retrieve_envflows(self):
        result = retrieve_env_interventions('sample_project', intervention_matrix='biosphere', keys="('biosphere', 'PM')")
        self.assertEqual(result[0]['name'], 'Particulate matter, industrial')

class TestPULPO(unittest.TestCase):

    def test_pulpo(self):
        project = 'sample_project'
        database = 'technosphere'
        methods = {"('my project', 'climate change')": 1,
                   "('my project', 'air quality')": 1,
                   "('my project', 'resources')": 0}
        # Test basic PULPO:
        worker = pulpo.PulpoOptimizer(project, database, methods, '')
        worker.intervention_matrix = 'biosphere'
        worker.get_lci_data()
        eCar = worker.retrieve_activities(reference_products='transport')
        demand = {eCar[0]: 1}
        elec = worker.retrieve_activities(reference_products='electricity')
        choices = {'electricity': {elec[0]: 100, elec[1]: 100}}
        worker.instantiate(choices=choices, demand=demand)
        worker.solve()
        result_obj = round(worker.instance.OBJ(), 6)
        result_aux = round(worker.instance.impacts_calculated["('my project', 'resources')"].value, 5)
        self.assertEqual(result_obj, 0.103093)
        self.assertEqual(result_aux, 5.25773)
        # Test supply specification:
        upper_limit = {eCar[0]: 1}
        lower_limit = {eCar[0]: 1}
        worker.instantiate(choices=choices, upper_limit=upper_limit, lower_limit=lower_limit)
        worker.solve()
        result_obj = round(worker.instance.OBJ(), 6)
        result_aux = round(worker.instance.impacts_calculated["('my project', 'resources')"].value, 5)
        self.assertEqual(result_obj, 0.1)
        self.assertEqual(result_aux, 5.1)
        # Test elementary / intervention flow constraint:
        water = worker.retrieve_envflows(activities="Water, irrigation")
        upper_elem_limit = {water[0]: 5.2}
        worker.instantiate(choices=choices, demand=demand, upper_elem_limit=upper_elem_limit)
        worker.solve()
        result_obj = round(worker.instance.OBJ(), 6)
        result_aux = round(worker.instance.impacts_calculated["('my project', 'resources')"].value, 5)
        self.assertEqual(result_obj, 0.14237)
        self.assertEqual(result_aux, 5.2)







import brightway2 as bw
import copy
import numpy as np
np.NaN = np.nan

# Helper function to convert integers to Latin letters
def int_to_latin(num):
    """Converts an integer to a Latin letter (e.g., 1 -> 'A', 2 -> 'B')."""
    return chr(64 + num)  # Converts 1 to 'A', 2 to 'B', etc.

# Helper function to convert integers to Greek letters
def int_to_greek(num):
    """Converts an integer to a Greek letter (e.g., 1 -> 'α', 2 -> 'β')."""
    greek_letters = ['α', 'β', 'γ', 'δ', 'ε', 'ζ', 'η', 'θ', 'ι', 'κ', 'λ']
    return greek_letters[num - 1] if num <= len(greek_letters) else str(num)

class Process:
    """Represents a production process for a specific product in a specific region."""

    def __init__(self, product_id, technology_id, region_id):
        """Initializes a process with product, technology, and region identifiers."""
        self.product_id = product_id
        self.technology_id = int_to_latin(technology_id)
        self.region_id = int_to_greek(region_id)
        self.process_id = f"{product_id}_{self.technology_id}_{self.region_id}"
        self.inputs = []  # List of tuples [(input_market_id, quantity), ...]
        self.environmental_flows = []  # List of tuples [(flow_id, quantity), ...]

    def add_input(self, input_market, quantity):
        """Adds an input market with a specified quantity to the process."""
        self.inputs.append((input_market.market_id, quantity))

    def add_environmental_flow(self, flow_id, quantity):
        """Adds an environmental flow with a specified quantity to the process."""
        self.environmental_flows.append((flow_id, quantity))

    def generate_inputs(self, potential_inputs, n_inputs):
        """Randomly generates and normalizes inputs for the process from potential markets."""
        if len(potential_inputs) < n_inputs:
            num_inputs = len(potential_inputs)
        else:
            num_inputs = np.random.randint(0, n_inputs + 1)

        if num_inputs > 0:
            selected_inputs = np.random.choice(potential_inputs, num_inputs, replace=False)

            # Generate random quantities for each selected input
            quantities = np.random.rand(num_inputs)

            # Normalize to ensure the sum is <= 1 (in order to have an M-matrix --- diagonal dominance)
            normalized_quantities = quantities / quantities.sum()

            # Scale down to ensure the sum is strictly smaller than 1
            scaling_factor = np.random.rand()  # A random factor between 0 and 1
            normalized_quantities *= scaling_factor

            for inp, quantity in zip(selected_inputs, normalized_quantities):
                self.add_input(inp, quantity)

    def generate_environmental_flows(self, flow_ids, n_flows):
        """Randomly generates environmental flows for the process."""
        num_flows = np.random.randint(1, n_flows + 1)
        selected_flows = np.random.choice(flow_ids, num_flows, replace=False)
        for flow_id in selected_flows:
            quantity = np.random.rand()
            self.add_environmental_flow(str(flow_id), quantity)

    def __repr__(self):
        return (f"Process(product_id={self.product_id}, "
                f"technology_id={self.technology_id}, "
                f"region_id={self.region_id}, "
                f"inputs={self.inputs}, "
                f"environmental_flows={self.environmental_flows})")

class Market:
    """Represents a market for a specific product in a specific region."""

    def __init__(self, product_id, region_id):
        """Initializes a market with product and region identifiers."""
        self.product_id = product_id
        self.region_id = int_to_greek(region_id)
        self.market_id = f"{product_id}_{self.region_id}"
        self.composition = []  # List of tuples [(process_id, share), ...]

    def add_process(self, process, share):
        """Adds a process with a specified share to the market composition."""
        self.composition.append((process.process_id, share))

    def generate_market_composition(self, candidate_processes):
        """Randomly generates a market composition from candidate processes."""
        included_processes = []
        for process in candidate_processes:
            if np.random.rand() < 0.75:
                share = np.random.rand()
                included_processes.append((process, share))

        # Ensure at least one process is included
        if not included_processes:
            process = np.random.choice(candidate_processes)
            share = np.random.rand()
            included_processes.append((process, share))

        for process, share in included_processes:
            self.add_process(process, share)

        # Normalize shares to sum to 1
        total_share = sum(share for _, share in self.composition)
        if total_share > 0:
            self.composition = [(pid, share / total_share) for pid, share in self.composition]

    def __repr__(self):
        return (f"Market(product_id={self.product_id}, "
                f"region_id={self.region_id}, "
                f"composition={self.composition})")

def create_system(n_prod, n_proc, n_reg, n_inputs, n_flows, seed=None):
    """
    Creates the technosphere and biosphere components for the system.

    Args:
        n_prod (int): Number of products.
        n_proc (int): Maximum number of processes per product.
        n_reg (int): Number of regions.
        n_inputs (int): Maximum number of inputs per process.
        n_flows (int): Number of environmental flows.
        seed (int, optional): Seed for reproducibility.

    Returns:
        processes (list): List of Process instances.
        markets (list): List of Market instances.
        flow_ids (list): List of environmental flow IDs.
    """
    if seed is not None:
        np.random.seed(seed)

    processes = []
    markets = []
    flow_ids = [f"e{i}" for i in range(1, n_flows + 1)]  # Environmental flow IDs: e1, e2, e3, ...

    # Step 1: Initialize Processes
    for product_id in range(1, n_prod + 1):
        regions = list(range(1, n_reg + 1))
        np.random.shuffle(regions)
        active_regions = regions[:np.random.randint(1, n_reg + 1)]

        for region_id in active_regions:
            technologies = list(range(1, n_proc + 1))
            np.random.shuffle(technologies)
            active_technologies = technologies[:np.random.randint(1, n_proc + 1)]

            for technology_id in active_technologies:
                proc = Process(product_id, technology_id, region_id)
                proc.generate_environmental_flows(flow_ids, n_flows)
                processes.append(proc)

    # Step 2: Generate Markets
    for product_id in range(1, n_prod + 1):
        regions = list(range(1, n_reg + 1))
        np.random.shuffle(regions)

        for region_id in regions:
            candidate_processes = [proc for proc in processes if proc.product_id == product_id and proc.region_id == int_to_greek(region_id)]
            if candidate_processes:
                market = Market(product_id, region_id)
                market.generate_market_composition(candidate_processes)
                markets.append(market)

    # Step 3: Assign Inputs to Processes
    for proc in processes:
        potential_inputs = [market for market in markets if market.region_id == proc.region_id]
        if potential_inputs:
            proc.generate_inputs(potential_inputs, n_inputs)

    return processes, markets, flow_ids

def assemble_technosphere_matrix(processes, markets):
    """
    Assembles the technosphere matrix from processes and markets.

    Args:
        processes (list): List of Process instances.
        markets (list): List of Market instances.

    Returns:
        np.ndarray: Technosphere matrix.
    """
    n = len(processes) + len(markets)
    A = np.zeros((n, n))

    process_index = {proc.process_id: idx for idx, proc in enumerate(processes)}
    market_index = {market.market_id: idx + len(processes) for idx, market in enumerate(markets)}

    # Fill in the technosphere matrix
    for proc in processes:
        i = process_index[proc.process_id]
        A[i, i] = 1  # Diagonal element for the process

        for market_id, qty in proc.inputs:
            j = market_index[market_id]
            A[j, i] = -qty  # Negative value for input

    for market in markets:
        i = market_index[market.market_id]
        A[i, i] = 1  # Diagonal element for the market

        for process_id, share in market.composition:
            j = process_index[process_id]
            A[j, i] = -share  # Market composition values

    return A

def assemble_biosphere_matrix(processes, flow_ids):
    """
    Assembles the biosphere matrix from processes and environmental flows.

    Args:
        processes (list): List of Process instances.
        flow_ids (list): List of environmental flow IDs.

    Returns:
        np.ndarray: Biosphere matrix.
    """
    n = len(processes) + len(flow_ids)
    B = np.zeros((len(flow_ids), len(processes)))

    process_index = {proc.process_id: idx for idx, proc in enumerate(processes)}
    flow_index = {flow_id: idx for idx, flow_id in enumerate(flow_ids)}

    # Fill in the biosphere matrix
    for proc in processes:
        j = process_index[proc.process_id]
        for flow_id, qty in proc.environmental_flows:
            i = flow_index[flow_id]
            B[i, j] = qty  # Positive value for environmental flow

    return B

class CharacterizationFactor:
    """Represents a characterization factor for an environmental flow under a specific method."""

    def __init__(self, method_id, flow_id, factor):
        """Initializes a characterization factor with a method, flow, and factor value."""
        self.method_id = method_id  # ID of the method (e.g., m1, m2, etc.)
        self.flow_id = flow_id      # ID of the environmental flow (e.g., e1, e2, etc.)
        self.factor = factor        # Characterization factor value

    def __repr__(self):
        return (f"CharacterizationFactor(method_id={self.method_id}, "
                f"flow_id={self.flow_id}, "
                f"factor={self.factor})")

def create_characterization_factors(n_methods, n_flows, seed=None):
    """
    Creates characterization factors for each method and environmental flow.

    Args:
        n_methods (int): Number of impact assessment methods.
        n_flows (int): Number of environmental flows.
        seed (int, optional): Seed for reproducibility.

    Returns:
        methods (list): List of method IDs.
        characterization_factors (list): List of CharacterizationFactor instances.
    """
    if seed is not None:
        np.random.seed(seed)

    methods = [f"m{i}" for i in range(1, n_methods + 1)]  # Method IDs: m1, m2, m3, ...
    characterization_factors = []

    for method_id in methods:
        num_factors = np.random.randint(1, n_flows + 1)  # Randomly determine the number of factors for this method
        selected_flows = np.random.choice([f"e{i}" for i in range(1, n_flows + 1)], num_factors, replace=False)
        for flow_id in selected_flows:
            factor = np.random.rand()  # Generate a random factor value
            characterization_factors.append(CharacterizationFactor(method_id, flow_id, factor))

    return methods, characterization_factors

def assemble_characterization_matrices(characterization_factors, flow_ids, methods):
    """
    Assembles characterization factor matrices for each method.

    Args:
        characterization_factors (list): List of CharacterizationFactor instances.
        flow_ids (list): List of environmental flow IDs.
        methods (list): List of method IDs.

    Returns:
        dict: Characterization matrices keyed by method ID.
    """
    n_methods = len(methods)
    n_flows = len(flow_ids)
    C_matrices = {method: np.zeros((n_flows, 1)) for method in methods}  # Initialize matrices

    flow_index = {flow_id: idx for idx, flow_id in enumerate(flow_ids)}

    for cf in characterization_factors:
        method_id = cf.method_id
        flow_id = cf.flow_id
        factor = cf.factor
        i = flow_index[flow_id]
        C_matrices[method_id][i, 0] = factor

    return C_matrices

def setup_biosphere_db(flow_ids):
    """Setup or update the biosphere database with environmental flows."""

    # Create biosphere database if it doesn't exist
    if 'biosphere3' not in bw.databases:
        biosphere_db = bw.Database('biosphere3')
        biosphere_data = {}

        # Define environmental flows
        for flow_id in flow_ids:
            biosphere_data[('biosphere3', flow_id)] = {
                'name': f'Environmental flow {flow_id}',
                'categories': ('environmental impact', ''),
                'type': 'emission',
                'unit': 'kg',
            }

        biosphere_db.write(biosphere_data)
        print('Biosphere database created')
    else:
        print('Biosphere database already exists')

def setup_technosphere_db(database, processes, markets):
    """Setup the technosphere database with generated processes and markets."""
    if 'generated_technosphere_db' not in bw.databases:
        technosphere_db = bw.Database(database)
        technosphere_db.write({})

        # Register each process
        for proc in processes:
            act = technosphere_db.new_activity(proc.process_id)
            act['unit'] = 'unit'  # Define the unit appropriately
            act['location'] = proc.region_id
            act['name'] = f'Process {proc.process_id}'
            act['reference product'] = f'Product {proc.product_id}'
            act.new_exchange(amount=1.0, input=act.key, type='production').save()
            act.save()

        # Register each market
        for market in markets:
            act = technosphere_db.new_activity(market.market_id)
            act['unit'] = 'unit'  # Define the unit appropriately
            act['location'] = market.region_id
            act['name'] = f'Market {market.market_id}'
            act['reference product'] = f'Product {market.product_id}'
            act.new_exchange(amount=1.0, input=act.key, type='production').save()
            act.save()

        print('Technosphere database created')
    else:
        print('Technosphere database already exists')

def add_exchanges_to_db(database, processes, markets):
    """Add exchanges to the technosphere and market activities in the database."""
    technosphere_db = bw.Database(database)
    biosphere_db = bw.Database('biosphere3')

    # Add exchanges for processes
    for proc in processes:
        act = technosphere_db.get(proc.process_id)

        # Add market inputs
        for market_id, qty in proc.inputs:
            market_act = technosphere_db.get(market_id)
            act.new_exchange(amount=qty, input=market_act.key, type='technosphere').save()

        # Add environmental flows (biosphere exchanges)
        for flow_id, qty in proc.environmental_flows:
            flow_act = biosphere_db.get(flow_id)
            act.new_exchange(amount=qty, input=flow_act.key, type='biosphere').save()

        act.save()

    # Add exchanges for markets
    for market in markets:
        act = technosphere_db.get(market.market_id)

        # Add process contributions to the market
        for process_id, share in market.composition:
            process_act = technosphere_db.get(process_id)
            act.new_exchange(amount=share, input=process_act.key, type='technosphere').save()

        act.save()

    print('Exchanges added to the technosphere and market activities in the database')

def setup_lcia_methods(characterization_factors, methods):
    """Setup LCIA methods based on generated characterization factors."""
    # Deregister existing methods
    methods_copy = copy.deepcopy(bw.methods)
    for method in methods_copy:
        bw.Method(method).deregister()

    # Register new LCIA methods
    for method_id in methods:
        method = bw.Method(('generated_system_example', method_id))
        method.register()

        cfs = []
        for cf in characterization_factors:
            if cf.method_id == method_id:
                flow_key = ('biosphere3', cf.flow_id)
                cfs.append((flow_key, cf.factor))

        method.write(cfs)

    print('LCIA methods and CFs defined')

def setup_generated_system(project, database, processes, markets, flow_ids, characterization_factors, methods):
    """Main function to set up the entire Brightway2 database."""
    bw.projects.set_current(project)
    setup_biosphere_db(flow_ids)
    setup_technosphere_db(database, processes, markets)
    add_exchanges_to_db(database, processes, markets)
    setup_lcia_methods(characterization_factors, methods)


import brightway2 as bw


def setup_generic_db(project, database, n_prod, n_proc, n_reg, n_inputs, n_flows, n_methods, seed=None,
                     return_data=False):
    """
    Sets up a generic LCI database in Brightway2 with specified parameters.

    Args:
        project (str): Name of the Brightway2 project to create or use.
        database (str): Name of the Brightway2 database to create or use.
        n_prod (int): Number of products to generate.
        n_proc (int): Maximum number of processes per product.
        n_reg (int): Number of regions where processes can be active.
        n_inputs (int): Maximum number of inputs per process.
        n_flows (int): Number of environmental flows to generate.
        n_methods (int): Number of impact assessment methods to create.
        seed (int, optional): Seed for reproducibility of random data generation.
        return_data (bool): If True, returns the generated matrices (technosphere, biosphere, and
            characterization).

    Returns:
        tuple: If `return_data` is True, returns a tuple containing:
            - technosphere_matrix (np.ndarray): The technosphere matrix.
            - biosphere_matrix (np.ndarray): The biosphere matrix.
            - characterization_matrices (dict): A dictionary of characterization factor matrices.
    """

    # Set the project, deleting existing databases if the project exists
    if project in bw.projects:
        bw.projects.set_current(project)
        print(f"Project '{project}' already exists. Deleting existing databases.")
        if database in bw.databases:
            del bw.databases[database]
        if 'biosphere3' in bw.databases:
            del bw.databases['biosphere3']
    else:
        bw.projects.set_current(project)
        print(f"Project '{project}' does not exist. Creating new project.")

    print("Proceeding with creating both technosphere and biosphere databases.")

    # Step 1: Create the technosphere and biosphere components
    processes, markets, flow_ids = create_system(n_prod, n_proc, n_reg, n_inputs, n_flows, seed=seed)

    # Step 2: Assemble the technosphere matrix
    technosphere_matrix = assemble_technosphere_matrix(processes, markets)

    # Step 3: Assemble the biosphere matrix
    biosphere_matrix = assemble_biosphere_matrix(processes, flow_ids)

    # Step 4: Create the characterization factors for impact assessment methods
    methods, characterization_factors = create_characterization_factors(n_methods, n_flows, seed=seed)

    # Step 5: Assemble the characterization factor matrices
    characterization_matrices = assemble_characterization_matrices(characterization_factors, flow_ids, methods)

    # Set up the Brightway2 database using the generated data
    setup_generated_system(project, database, processes, markets, flow_ids, characterization_factors, methods)

    if return_data:
        return technosphere_matrix, biosphere_matrix, characterization_matrices


def main():
    setup_generic_db(project="generic_db_project", database="generic_db", n_prod=5, n_proc=3, n_reg=3, n_inputs=4, n_flows=4, n_methods=2, seed=None, return_data=False)

if __name__ == '__main__':
    main()



import pyomo.environ as pyo
from pyomo.contrib import appsi

def create_model():
    """
    Builds an abstract model on top of the ecoinvent database.

    Returns:
        AbstractModel: The Pyomo abstract model for optimization.
    """
    model = pyo.AbstractModel()

    # Sets
    model.PRODUCT = pyo.Set(doc='Set of intermediate products (or technosphere exchanges), indexed by i')
    model.PROCESS = pyo.Set(doc='Set of processes (or activities), indexed by j')
    model.ENV_COST = pyo.Set(doc='Set of environmental cost flows, indexed by e')
    model.INDICATOR = pyo.Set(doc='Set of impact assessment indicators, indexed by h')
    model.INV = pyo.Set(doc='Set of intervention flows, indexed by g')
    model.ENV_COST_PROCESS = pyo.Set(within=model.ENV_COST * model.PROCESS * model.INDICATOR, doc='Relation set between environmental cost flows and processes')
    model.ENV_COST_IN = pyo.Set(model.INDICATOR, within=model.ENV_COST)
    model.ENV_COST_OUT = pyo.Set(model.ENV_COST * model.INDICATOR, within=model.PROCESS)
    model.PROCESS_IN = pyo.Set(model.PROCESS, within=model.PRODUCT)
    model.PROCESS_OUT = pyo.Set(model.PRODUCT, within=model.PROCESS)
    model.PRODUCT_PROCESS = pyo.Set(within=model.PRODUCT * model.PROCESS, doc='Relation set between intermediate products and processes')
    model.INV_PROCESS = pyo.Set(within=model.INV * model.PROCESS, doc='Relation set between environmental flows and processes')
    model.INV_OUT = pyo.Set(model.INV, within=model.PROCESS)

    # Parameters
    model.UPPER_LIMIT = pyo.Param(model.PROCESS, mutable=True, within=pyo.Reals, doc='Maximum production capacity of process j')
    model.LOWER_LIMIT = pyo.Param(model.PROCESS, mutable=True, within=pyo.Reals, doc='Minimum production capacity of process j')
    model.UPPER_INV_LIMIT = pyo.Param(model.INV, mutable=True, within=pyo.Reals, doc='Maximum intervention flow g')
    model.UPPER_IMP_LIMIT = pyo.Param(model.INDICATOR, mutable=True, within=pyo.Reals, doc='Maximum impact on category h')
    model.ENV_COST_MATRIX = pyo.Param(model.ENV_COST_PROCESS, mutable=True, doc='Enviornmental cost matrix Q*B describing the environmental cost flows e associated to process j')
    model.INV_MATRIX = pyo.Param(model.INV_PROCESS, mutable=True, doc='Intervention matrix B describing the intervention flow g entering/leaving process j')
    model.FINAL_DEMAND = pyo.Param(model.PRODUCT, mutable=True, within=pyo.Reals, doc='Final demand of intermediate product flows (i.e., functional unit)')
    model.SUPPLY = pyo.Param(model.PRODUCT, mutable=True, within=pyo.Binary, doc='Binary parameter which specifies whether or not a supply has been specified instead of a demand')
    model.TECH_MATRIX = pyo.Param(model.PRODUCT_PROCESS, mutable=True, doc='Technology matrix A describing the intermediate product i produced/absorbed by process j')
    model.WEIGHTS = pyo.Param(model.INDICATOR, mutable=True, within=pyo.NonNegativeReals, doc='Weighting factors for the impact assessment indicators in the objective function')

    # Variables
    model.impacts = pyo.Var(model.INDICATOR, bounds=(-1e24, 1e24), doc='Environmental impact on indicator h evaluated with the established LCIA method')
    model.scaling_vector = pyo.Var(model.PROCESS, bounds=(-1e24, 1e24), doc='Activity level of each process to meet the final demand')
    model.inv_vector = pyo.Var(model.INV, bounds=(-1e24, 1e24), doc='Intervention flows')
    model.slack = pyo.Var(model.PRODUCT, bounds=(0, 1e24), doc='Supply slack variables')

    # Building rules for sets
    model.Env_in_out = pyo.BuildAction(rule=populate_env)
    model.Process_in_out = pyo.BuildAction(rule=populate_in_and_out)
    model.Inv_in_out = pyo.BuildAction(rule=populate_inv)

    # Constraints
    model.FINAL_DEMAND_CNSTR = pyo.Constraint(model.PRODUCT, rule=demand_constraint)
    model.IMPACTS_CNSTR = pyo.Constraint(model.INDICATOR, rule=impact_constraint)
    model.INVENTORY_CNSTR = pyo.Constraint(model.INV, rule=inventory_constraint)
    model.UPPER_CNSTR = pyo.Constraint(model.PROCESS, rule=upper_constraint)
    model.LOWER_CNSTR = pyo.Constraint(model.PROCESS, rule=lower_constraint)
    model.SLACK_CNSTR = pyo.Constraint(model.PRODUCT, rule=slack_constraint)
    model.INV_CNSTR = pyo.Constraint(model.INV, rule=upper_env_constraint)
    model.IMP_CNSTR = pyo.Constraint(model.INDICATOR, rule=upper_imp_constraint)

    # Objective function
    model.OBJ = pyo.Objective(sense=pyo.minimize, rule=objective_function)

    return model


# Rule functions
def populate_env(model):
    """Relates the environmental flows to the processes."""
    for i, j, h in model.ENV_COST_PROCESS:
        if i not in model.ENV_COST_IN[h]:
            model.ENV_COST_IN[h].add(i)
        model.ENV_COST_OUT[i, h].add(j)


def populate_in_and_out(model):
    """Relates the inputs of an activity to its outputs."""
    for i, j in model.PRODUCT_PROCESS:
        model.PROCESS_OUT[i].add(j)
        model.PROCESS_IN[j].add(i)

def populate_inv(model):
    """Relates the impacts to the environmental flows"""
    for a, j in model.INV_PROCESS:
        model.INV_OUT[a].add(j)

def demand_constraint(model, i):
    """Fixes a value in the demand vector"""
    return sum(model.TECH_MATRIX[i, j] * model.scaling_vector[j] for j in model.PROCESS_OUT[i]) == model.FINAL_DEMAND[i] + model.slack[i]

def impact_constraint(model, h):
    """Calculates all the impact categories"""
    return model.impacts[h] == sum(sum(model.ENV_COST_MATRIX[i, j, h] * model.scaling_vector[j] for j in model.ENV_COST_OUT[i, h]) for i in model.ENV_COST_IN[h])

def inventory_constraint(model, g):
    """Calculates the environmental flows"""
    return model.inv_vector[g] == sum(model.INV_MATRIX[g, j] * model.scaling_vector[j] for j in model.INV_OUT[g])

def upper_constraint(model, j):
    """Ensures that variables are within capacities (Maximum production constraint) """
    return model.scaling_vector[j] <= model.UPPER_LIMIT[j]

def lower_constraint(model, j):
    """ Minimum production constraint """
    return model.scaling_vector[j] >= model.LOWER_LIMIT[j]

def upper_env_constraint(model, g):
    """Ensures that variables are within capacities (Maximum production constraint) """
    return model.inv_vector[g] <= model.UPPER_INV_LIMIT[g]

def upper_imp_constraint(model, h):
    """ Imposes upper limits on selected impact categories """
    return model.impacts[h] <= model.UPPER_IMP_LIMIT[h]


def slack_constraint(model, j):
    """ Slack variable upper limit for activities where supply is specified instead of demand """
    return model.slack[j] <= 1e20 * model.SUPPLY[j]

def objective_function(model):
    """Objective is a sum over all indicators with weights. Typically, the indicator of study has weight 1, the rest 0"""
    return sum(model.impacts[h] * model.WEIGHTS[h] for h in model.INDICATOR)

def calculate_methods(instance, lci_data, methods):
    """
    Calculates the impacts if a method with weight 0 has been specified.

    Args:
        instance: The Pyomo model instance.
        lci_data (dict): LCI data containing matrices and mappings.
        methods (dict): Methods for environmental impact assessment.

    Returns:
        instance: The updated Pyomo model instance with calculated impacts.
    """
    import scipy.sparse as sparse
    import numpy as np
    matrices = lci_data['matrices']
    intervention_matrix = lci_data['intervention_matrix']
    matrices = {h: matrices[h] for h in matrices if str(h) in methods}
    env_cost = {h: sparse.csr_matrix.dot(matrices[h], intervention_matrix) for h in matrices}
    try:
        scaling_vector = np.array([instance.scaling_vector[x].value for x in instance.scaling_vector])
    except:
        scaling_vector = np.array([instance.scaling_vector[x] for x in instance.scaling_vector])
    impacts = {}
    for h in matrices:
        impacts[h] = sum(env_cost[h].dot(scaling_vector))
    instance.impacts_calculated = pyo.Var([h for h in impacts], initialize=impacts)
    return instance

def calculate_inv_flows(instance, lci_data):
    """
    Calculates elementary flows post-optimization.

    Args:
        instance: The Pyomo model instance.
        lci_data (dict): LCI data containing matrices and mappings.

    Returns:
        instance: The updated Pyomo model instance with calculated intervention flows.
    """
    import numpy as np
    intervention_matrix = lci_data['intervention_matrix']
    try:
        scaling_vector = np.array([instance.scaling_vector[x].value for x in instance.scaling_vector])
    except:
        scaling_vector = np.array([instance.scaling_vector[x] for x in instance.scaling_vector])
    flows = intervention_matrix.dot(scaling_vector)
    instance.inv_flows = pyo.Var(range(0, intervention_matrix.shape[0]), initialize=flows)
    return instance



def instantiate(model_data):
    """
    Builds an instance of the optimization model with specific data and objective function.

    Args:
        model_data (dict): Data dictionary for the optimization model.

    Returns:
        ConcreteModel: The instantiated Pyomo model.
    """
    print('Creating Instance')
    model = create_model()
    problem = model.create_instance(model_data, report_timing=False)
    print('Instance created')
    return problem


def solve_model(model_instance, gams_path=None, solver_name=None, options=None):
    """
    Solves the instance of the optimization model.

    Args:
        model_instance (ConcreteModel): The Pyomo model instance.
        gams_path (str, optional): Path to the GAMS solver. If None, GAMS will not be used.
        solver_name (str, optional): The solver to use ('highs', 'gams', or 'ipopt'). Defaults to 'highs' unless gams_path is provided.
        options (list, optional): Additional options for the solver.

    Returns:
        tuple: Results of the optimization and the updated model instance.
    """
    results = None

    # Use GAMS if gams_path is specified
    if gams_path and (solver_name is None or solver_name.lower() == 'gams'):
        pyo.pyomo.common.Executable('gams').set_path(gams_path)
        solver = pyo.SolverFactory('gams')
        print('GAMS solvers library availability:', solver.available())
        print('Solver path:', solver.executable())

        io_options = {
            'mtype': 'lp',  # Type of problem (lp, nlp, mip, minlp)
            'solver': 'CPLEX',  # Name of solver
        }

        if options is None:
            options = [
                'option optcr = 1e-15;',
                'option reslim = 3600;',
                'GAMS_MODEL.optfile = 1;',
                '$onecho > cplex.opt',
                'workmem=4096',
                'scaind=1',
                #'numericalemphasis=1',
                #'epmrk=0.99',
                #'eprhs=1E-9',
                '$offecho',
            ]

        results = solver.solve(
            model_instance,
            keepfiles=True,
            symbolic_solver_labels=True,
            tee=False,
            report_timing=False,
            io_options=io_options,
            add_options=options
        )

        model_instance.solutions.load_from(results)

    # Use IPOPT if explicitly specified
    elif solver_name and solver_name.lower() == 'ipopt':
        opt = pyo.SolverFactory('ipopt')
        if options:
            for option in options:
                opt.options[option] = True
        results = opt.solve(model_instance)

    # Default to HiGHS if no solver specified or if solver_name is 'highs'
    else:
        opt = appsi.solvers.Highs()
        results = opt.solve(model_instance)

    return results, model_instance


    print('Optimization problem solved')
    ## TODO: Add a check for infeasibility and other solver errors

    return results, model_instance


import scipy.sparse as sparse


def combine_inputs(lci_data, demand, choices, upper_limit, lower_limit, upper_inv_limit, upper_imp_limit, methods):
    """
    Combines all the inputs into a dictionary as an input for the optimization model.

    Args:
        lci_data (dict): LCI data containing matrices and mappings.
        demand (dict): Demand data.
        choices (dict): Choices for the model.
        upper_limit (dict): Upper limit constraints.
        lower_limit (dict): Lower limit constraints.
        upper_inv_limit (dict): Upper intervention limit constraints.
        upper_imp_limit (dict): Upper impact limit constraints.
        methods (dict): Methods for environmental impact assessment.

    Returns:
        dict: Combined data dictionary for the optimization model.
    """

    # Load LCI data matrices and mappings
    matrices = lci_data['matrices']
    intervention_matrix = lci_data['intervention_matrix']
    technology_matrix = lci_data['technology_matrix']
    process_map = lci_data['process_map']
    intervention_map = lci_data['intervention_map']

    # Remove matrices that are not part of the objective
    matrices = {h: matrices[h] for h in matrices if str(h) in methods}

    # Prepare the environmental impact matrices Q[h]*B
    env_cost = {h: sparse.csr_matrix.dot(matrices[h], intervention_matrix) for h in matrices}

    # Convert sparse csr biosphere impact matrix to dictionary
    env_cost_dict = {(i - 1, env_cost[h].indices[j], h): env_cost[h].data[j]
                     for h in matrices for i in range(1, intervention_matrix.shape[0] + 1)
                     for j in range(env_cost[h].indptr[i - 1], env_cost[h].indptr[i])}

    # Convert sparse csr technology matrix to dictionary
    technology_matrix_dict = {(i - 1, technology_matrix.indices[j]): technology_matrix.data[j]
                              for i in range(1, technology_matrix.shape[0] + 1)
                              for j in range(technology_matrix.indptr[i - 1], technology_matrix.indptr[i])}

    # Convert sparse csr intervention flow matrix to dictionary
    inv_to_consider = [intervention_map[g] for g in upper_inv_limit]
    inv_dict = {(g, intervention_matrix.indices[j]): intervention_matrix.data[j]
                for g in inv_to_consider
                for j in range(intervention_matrix.indptr[g], intervention_matrix.indptr[g + 1])}

    # Make technology matrix rectangular and update keys and product_ids
    keys = {}  # Store the activity keys of the choices
    product_ids = []  # Store the product IDs of the choices
    for key, processes in choices.items():
        for proc in processes:
            product_id = process_map[proc.key]
            keys[product_id] = key
            product_ids.append(product_id)

    for product, process in list(technology_matrix_dict):
        if product in product_ids:
            if (keys[product], process) not in technology_matrix_dict:
                technology_matrix_dict[(keys[product], process)] = technology_matrix_dict[(product, process)]
            else:
                technology_matrix_dict[(keys[product], process)] += technology_matrix_dict[(product, process)]
            del technology_matrix_dict[(product, process)]

    # Create sets using set comprehensions for better performance
    PRODUCTS = {None: list({i[0] for i in technology_matrix_dict})}
    PROCESS = {None: list({i[1] for i in technology_matrix_dict})}
    PRODUCT_PROCESS = {None: list({(i[0], i[1]) for i in technology_matrix_dict})}
    ENV_COST = {None: list({i[0] for i in env_cost_dict})}
    ENV_COST_PROCESS = {None: list({i for i in env_cost_dict})}
    INV = {None: list({i[0] for i in inv_dict})}
    INV_PROCESS = {None: list({(i[0], i[1]) for i in inv_dict})}
    INDICATOR = {None: list({h for h in matrices})}

    # Specify the demand
    demand_dict = {prod: 0 for prod in PRODUCTS[None]}
    for dem in demand:
        if dem in process_map:
            # Case when dem is a key in process_map
            demand_dict[process_map[dem]] = demand[dem]
        elif dem in choices:
            # Case when dem is already one of the elements in process_map
            demand_dict[dem] = demand[dem]
        else:
            # Case when dem is neither a key nor a value in process_map
            raise ValueError(f"'{dem}' is not found in process_map keys or values.")

    # Specify the lower limit
    lower_limit_dict = {proc: -1e20 for proc in PROCESS[None]}
    for choice in choices:
        for proc in choices[choice]:
            lower_limit_dict[process_map[proc]] = 0
    for proc in lower_limit:
        lower_limit_dict[process_map[proc]] = lower_limit[proc]

    # Specify the upper limit
    upper_limit_dict = {proc: 1e20 for proc in PROCESS[None]}
    for proc in upper_limit:
        upper_limit_dict[process_map[proc]] = upper_limit[proc]
    for choice in choices:
        for proc in choices[choice]:
            upper_limit_dict[process_map[proc]] = choices[choice][proc]

    # Check if a supply has been specified
    supply_dict = {prod: 0 for prod in PRODUCTS[None]}
    for proc in list(lower_limit.keys() & upper_limit.keys()):
        supply_dict[process_map[proc]] = 1 if lower_limit[proc] == upper_limit[proc] else 0

    # Specify the upper elementary flow limit
    upper_inv_limit_dict = {elem: 1e24 for elem in INV[None]}
    for inv in upper_inv_limit:
        upper_inv_limit_dict[intervention_map[inv.key]] = upper_inv_limit[inv]

    # Specify the upper impact category limit
    upper_imp_limit_dict = {imp: 1e24 for imp in INDICATOR[None]}
    for imp in upper_imp_limit:
        upper_imp_limit_dict[imp] = upper_imp_limit[imp]

    # Create weights
    weights = {method: 1 for method in matrices} if methods == {} else methods

    # Assemble the final data dictionary
    model_data = {
        None: {
            'PRODUCT': PRODUCTS,
            'PROCESS': PROCESS,
            'ENV_COST': ENV_COST,
            'INDICATOR': INDICATOR,
            'INV': INV,
            'PRODUCT_PROCESS': PRODUCT_PROCESS,
            'ENV_COST_PROCESS': ENV_COST_PROCESS,
            'INV_PROCESS': INV_PROCESS,
            'TECH_MATRIX': technology_matrix_dict,
            'ENV_COST_MATRIX': env_cost_dict,
            'INV_MATRIX': inv_dict,
            'FINAL_DEMAND': demand_dict,
            'SUPPLY': supply_dict,
            'LOWER_LIMIT': lower_limit_dict,
            'UPPER_LIMIT': upper_limit_dict,
            'UPPER_INV_LIMIT': upper_inv_limit_dict,
            'UPPER_IMP_LIMIT': upper_imp_limit_dict,
            'WEIGHTS': weights,
        }
    }
    return model_data


def convert_to_dict(input_data):
    """
    Checks if the passed method is in the correct format and converts it accordingly.

    Args:
        input_data (Union[str, list, dict]): Input data to be converted.

    Returns:
        dict: Converted dictionary with methods and their weights.

    Raises:
        ValueError: If the input data is not in the expected format.
    """

    # Check if the input is a string
    if isinstance(input_data, str):
        return {input_data: 1}

    # Check if the input is a list of strings
    elif isinstance(input_data, list) and all(isinstance(item, str) for item in input_data):
        return {item: 1 for item in input_data}

    # Check if the input is a dictionary with numerical values
    elif isinstance(input_data, dict) and all(isinstance(value, (int, float)) for value in input_data.values()):
        return input_data

    # If the input is of any other type, raise an error
    else:
        raise ValueError(
            "Input should be either a string, a list of strings, or a dictionary with string indices and numerical values")



import pyomo.environ as pyo
from pyomo.repn.plugins.baron_writer import *
import pandas as pd
from pathlib import Path
from IPython.display import display

def save_results(instance, project, database, choices, constraints, demand, process_map, itervention_map, directory, name):
    """
    Saves the results of a Pyomo optimization model to an Excel file.

    Args:
        instance: The Pyomo model instance.
        project (str): Name of the project.
        database (str): Name of the database.
        choices (dict): Choices for the model.
        constraints (dict): Constraints applied during optimization.
        demand (dict): Demand data used in optimization.
        process_map (dict): Mapping of process IDs to descriptions.
        itervention_map (dict): Mapping of intervention IDs to descriptions.
        directory (str): Directory to save the results file.
        name (str): Name of the results file.
    """
    # Check if data/results folder exists, if not create it
    Path(directory + '/results').mkdir(parents=True, exist_ok=True)

    # Recover dictionary values
    list_of_vars = []
    for v in instance.component_objects(ctype=pyo.Var, active=True, descend_into=True):
        for e in v._data:
            v._data[e] = value(v[e])
        list_of_vars.append(v)

    # Open xlsx writer
    writer = pd.ExcelWriter(directory + '/results/' + name, engine='xlsxwriter')

    # Raw results
    for v in list_of_vars:
        try:
            if str(v) == 'inv_flows' or str(v) == 'inv_vector':
                data = [(k, itervention_map[k], v) for k, v in v._data.items()]
            else:
                data = [(k, process_map[k], v) for k, v in v._data.items()]
            df = pd.DataFrame(data, columns=['ID', 'Process', 'Value'])
        except:
            data = [(k, v) for k, v in v._data.items()]
            df = pd.DataFrame(data, columns=['Key', 'Value'])
        df.sort_values(by=['Value'], inplace=True, ascending=False)
        df.to_excel(writer, sheet_name=v.name, index=False)

    # Store the metadata
    pd.DataFrame([project + '__' + database]).to_excel(writer, sheet_name='project and db')

    metadata = {}
    for choice in choices:
        i = 0
        temp_dict = []
        for alt in choices[choice]:
            temp_dict.append((alt, i, instance.scaling_vector[process_map[alt.key]]))
            i+=1
        metadata[(choice, 'Process')] = {'Process ' + str(i): process_map[process_map[alt.key]] for alt, i, val in temp_dict}
        metadata[(choice, 'Capacity')] = {'Process ' + str(i): choices[choice][alt] for alt, i, val in temp_dict}
        metadata[(choice, 'Value')] = {'Process ' + str(i): x for alt, i, x in temp_dict}

    pd.DataFrame(metadata).to_excel(writer, sheet_name='choices')

    metadata = {}
    metadata['Demand'] = {
        process_map[process_map[key]] if key in process_map else key: demand[key]
        for key in demand
    }
    pd.DataFrame(metadata).to_excel(writer, sheet_name='demand')

    metadata = {}
    metadata['Demand'] = {process_map[process_map[key]]: constraints[key] for key in constraints}
    pd.DataFrame(metadata).to_excel(writer, sheet_name='constraints')

    # Save xlsx file
    writer.close()
    return

def summarize_results(instance, choices, constraints, demand, process_map, zeroes):
    """
    Summarizes the results of the optimization and prints them to the console.

    Args:
        instance: The Pyomo model instance.
        choices (dict): Choices for the model.
        constraints (dict): Constraints applied during optimization.
        demand (dict): Demand data used in optimization.
        process_map (dict): Mapping of process IDs to descriptions.
        zeroes (bool): Whether to include zero values in the summary.
    """
    metadata = {}
    metadata['Demand'] = {
        process_map[process_map[key]] if key in process_map else key: demand[key]
        for key in demand
    }

    print('The following demand / functional unit has been specified: ')
    display(pd.DataFrame(metadata))

    # Recover dictionary values
    list_of_vars = []
    for v in instance.component_objects(ctype=pyo.Var, active=True, descend_into=True):
        for e in v._data:
            v._data[e] = value(v[e])
        list_of_vars.append(v)

    # Raw results
    for v in list_of_vars:
        if v.name == 'impacts':
            try:
                data = [(k, process_map[k], v) for k, v in v._data.items()]
                df = pd.DataFrame(data, columns=['ID', 'Process', 'Value'])
            except:
                data = [(k, v) for k, v in v._data.items()]
                df = pd.DataFrame(data, columns=['Key', 'Value'])
            df.sort_values(by=['Value'], inplace=True, ascending=False)
            print('\nThese are the impacts contained in the objective:')
            display(df)

        if v.name == 'impacts_calculated':
            try:
                data = [(k, process_map[k], v) for k, v in v._data.items()]
                df = pd.DataFrame(data, columns=['ID', 'Process', 'Value'])
            except:
                data = [(k, v) for k, v in v._data.items()]
                df = pd.DataFrame(data, columns=['Key', 'Value'])
            df.sort_values(by=['Value'], inplace=True, ascending=False)
            print('\nThe following impacts were calculated: ')
            display(df)

    print('\nThe following choices were made: ')
    for choice in choices:
        metadata = {}
        i = 0
        temp_dict = []
        for alt in choices[choice]:
            if zeroes == False or instance.scaling_vector[process_map[alt.key]] != 0:
                temp_dict.append((alt, i, instance.scaling_vector[process_map[alt.key]]))
                i += 1
        metadata[(choice, 'Process')] = {'Process ' + str(i): process_map[process_map[alt.key]] for alt, i, val in temp_dict}
        metadata[(choice, 'Capacity')] = {'Process ' + str(i): choices[choice][alt] for alt, i, val in temp_dict}
        metadata[(choice, 'Value')] = {'Process ' + str(i): x for alt, i, x in temp_dict}
        print(choice)
        display(pd.DataFrame(metadata))

    if constraints == {}:
        print('No additional constraints have been passed.')
    else:
        metadata = {}
        metadata['Constraints'] = {process_map[process_map[key]]: constraints[key] for key in constraints}
        print('\nThe following constraints were implemented and oblieged: ')
        display(pd.DataFrame(metadata))

